/*
 * Copyright (C) 2013 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.build.gradle.internal.dsl;

import com.android.annotations.NonNull;
import com.android.annotations.Nullable;
import com.google.common.base.Objects;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;

import org.gradle.api.tasks.Input;
import org.gradle.api.tasks.Optional;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;

/**
 * DSL object for per-variant NDK settings, such as the ABI filter.
 */
public class NdkOptions implements CoreNdkOptions, Serializable {
    private static final long serialVersionUID = 1L;

    private String moduleName;
    private String cFlags;
    private List<String> ldLibs;
    private Set<String> abiFilters;
    private String stl;
    private Integer jobs;

    public NdkOptions() {
    }

    public void _initWith(@NonNull CoreNdkOptions ndkConfig) {
        moduleName = ndkConfig.getModuleName();
        cFlags = ndkConfig.getcFlags();
        setLdLibs(ndkConfig.getLdLibs());
        setAbiFilters(ndkConfig.getAbiFilters());
    }

    @Override
    @Input @Optional
    public String getModuleName() {
        return moduleName;
    }

    public void setModuleName(String moduleName) {
        this.moduleName = moduleName;
    }

    @Override
    @Input @Optional
    public String getcFlags() {
        return cFlags;
    }

    public void setcFlags(String cFlags) {
        this.cFlags = cFlags;
    }

    @Override
    @Input @Optional
    public List<String> getLdLibs() {
        return ldLibs;
    }

    @NonNull
    public NdkOptions ldLibs(String lib) {
        if (ldLibs == null) {
            ldLibs = Lists.newArrayList();
        }
        ldLibs.add(lib);
        return this;
    }

    @NonNull
    public NdkOptions ldLibs(String... libs) {
        if (ldLibs == null) {
            ldLibs = Lists.newArrayListWithCapacity(libs.length);
        }
        Collections.addAll(ldLibs, libs);
        return this;
    }

    @NonNull
    public NdkOptions setLdLibs(Collection<String> libs) {
        if (libs != null) {
            if (ldLibs == null) {
                ldLibs = Lists.newArrayListWithCapacity(libs.size());
            } else {
                ldLibs.clear();
            }
            ldLibs.addAll(libs);
        } else {
            ldLibs = null;
        }
        return this;
    }

    /**
     * ABI configurations of your native libraries Gradle should
     * build and package with your APK. You can list any subset of the
     * <a href="https://developer.android.com/ndk/guides/abis.html#sa">
     * ABIs the NDK supports</a>.
     * <p>For example:</p>
     * <p><code>abiFilters 'x86', 'x86_64', 'armeabi'</code></p>
     *
     * <p>When this flag is not configured, Gradle builds and packages all available ABIs.</p>
     */
    @Override
    @Input @Optional
    public Set<String> getAbiFilters() {
        return abiFilters;
    }


    @NonNull
    public NdkOptions abiFilter(String filter) {
        if (abiFilters == null) {
            abiFilters = Sets.newHashSetWithExpectedSize(2);
        }
        abiFilters.add(filter);
        return this;
    }

    @NonNull
    public NdkOptions abiFilters(String... filters) {
        if (abiFilters == null) {
            abiFilters = Sets.newHashSetWithExpectedSize(2);
        }
        Collections.addAll(abiFilters, filters);
        return this;
    }

    @NonNull
    public NdkOptions setAbiFilters(Collection<String> filters) {
        if (filters != null) {
            if (abiFilters == null) {
                abiFilters = Sets.newHashSetWithExpectedSize(filters.size());
            } else {
                abiFilters.clear();
            }
            abiFilters.addAll(filters);
        } else {
            abiFilters = null;
        }
        return this;
    }

    @Override
    @Nullable
    public String getStl() {
        return stl;
    }

    public void setStl(String stl) {
        this.stl = stl;
    }

    @Nullable
    @Override
    public Integer getJobs() {
        return jobs;
    }

    public void setJobs(Integer jobs) {
        this.jobs = jobs;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }
        NdkOptions that = (NdkOptions) o;
        return Objects.equal(moduleName, that.moduleName) &&
                Objects.equal(cFlags, that.cFlags) &&
                Objects.equal(ldLibs, that.ldLibs) &&
                Objects.equal(abiFilters, that.abiFilters) &&
                Objects.equal(stl, that.stl) &&
                Objects.equal(jobs, that.jobs);
    }

    @Override
    public int hashCode() {
        return Objects.hashCode(moduleName, cFlags, ldLibs, abiFilters, stl, jobs);
    }
}
